import logging
from .settings import *
# import re
# from time import gmtime, strftime

# COLORIZE_LOGS = os.getenv('DSO_COLORIZE_LOGS') or True
# BOLD_LOGS = os.getenv('DSO_BOLD_LOGS') or True
# TIMESTAMP_LOGS = os.getenv('DSO_TIMESTAMP_LOGS') or True
# LABEL_LOG_LEVELS = os.getenv('DSO_LABEL_LOG_LEVELS') or True

COLOR_CODES = {
    'black': u'0;30', 
    'bright gray': u'0;37',
    'blue': u'0;34', 
    'white': u'1;37',
    'green': u'0;32', 
    'bright blue': u'1;34',
    'cyan': u'0;36', 
    'bright green': u'1;32',
    'red': u'0;31', 
    'bright cyan': u'1;36',
    'purple': u'0;35', 
    'bright red': u'1;31',
    'yellow': u'0;33', 
    'bright purple': u'1;35',
    'dark gray': u'1;30', 
    'bright yellow': u'1;33',
    'magenta': u'0;35', 
    'bright magenta': u'1;35',
    'normal': u'0',
}

log_levels = {
    'critical': 0, ### logs only fatal errors
    'error': 1, ### logs only errors
    'warning': 2, ### logs also warnings
    'info': 3, ### logs also information
    'debug': 4, ### logs also debug information
    'exception': 5, ### logs also unhandled exceptions
    'packages': 6, ### also sets built-in logger level to logging.INFo
    'full': 7, ### also sets built-in logger level to logging.DEBUG
}

logger_verbosity_map = {
    '0': logging.CRITICAL,
    '1': logging.ERROR,
    '2': logging.WARNING,
    '3': logging.WARNING,
    '4': logging.WARNING,
    '5': logging.WARNING,
    '6': logging.INFO,
    '7': logging.DEBUG,
}

style_codes = {
    'normal': '\033[0m',
    'bold': '\033[1m',
    'reset': '\033[0m',
}

color_codes = {
    'uncolored': {
        'normal': '\033[0m',
        'bold': '\033[1m',
    },
    'white': {
        'normal': '\033[{0}m'.format(COLOR_CODES['white']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['white']),
    },
    'green': {
        'normal': '\033[{0}m'.format(COLOR_CODES['green']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['green']),
    },
    'yellow': {
        'normal': '\033[{0}m'.format(COLOR_CODES['yellow']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['yellow']),
    },
    'red': {
        'normal': '\033[{0}m'.format(COLOR_CODES['red']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['red']),
    },
    'debug': {
        'normal': '\033[{0}m'.format(COLOR_CODES['dark gray']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['dark gray']),
    },
    'info': {
        'normal': '\033[{0}m'.format(COLOR_CODES['white']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['white']),
    },
    'warning': {
        'normal': '\033[{0}m'.format(COLOR_CODES['yellow']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['yellow']),
    },
    'error': {
        'normal': '\033[{0}m'.format(COLOR_CODES['red']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['red']),
    },
    'critical': {
        'normal': '\033[{0}m'.format(COLOR_CODES['red']),
        'bold': '\033[{0}m\033[1m'.format(COLOR_CODES['red']),
    },
}

# 
# 

# def decode_color_codes(s):
#     r = re.findall("{0}(.*?){1}(.*?){2}".format(re.escape('<?'), re.escape('>'), re.escape('</?>')), s)
#     for c in r:
#         d = format_text(c[1], c[0])
#         s = s.replace('<?{0}>{1}</?>'.format(c[0], c[1]), d)
#     return s






def format_text(msg, color):
    """Given a string add necessary codes to format the string."""
    style = 'bold' if BOLD_LOGS else 'normal'
    if COLORIZE_LOGS:
        return '{0}{1}{2}'.format(color_codes[color][style], msg, style_codes['reset'])
    else:
        return '{0}{1}{2}'.format(color_codes['uncolored'][style], msg, style_codes['reset'])





class LoggerClass():

    @property
    def level(self):
        return int(self.__level)

    @property
    def mapped_log_level(self):
        return logger_verbosity_map[str(self.__level)]

    def __init__(self, log_level):
        self.__level = log_level
        log_format = ''
        if TIMESTAMP_LOGS: log_format += '%(asctime)s'
        if LABEL_LOG_LEVELS: log_format += ' [%(levelname)-8s]'
        log_format += ' %(message)s'
        logging.basicConfig(format=log_format, datefmt='%Y-%m-%d %H:%M:%S', level=self.mapped_log_level)
        self.LOG = logging.getLogger('dsocli')

    def set_verbosity(self, log_level):
        self.__level = log_level
        logging.basicConfig(level=self.mapped_log_level)

    def critical(self, msg, force=True):
        if force or self.level >= log_levels['critical']: 
            saveLevel = self.LOG.root.level
            try:
                self.LOG.root.level=logging.DEBUG
                self.LOG.critical(format_text(msg, 'critical'))
            finally:
                self.LOG.root.level=saveLevel

    def error(self, msg, force=True):
        if force or self.level >= log_levels['error']: 
            saveLevel = self.LOG.root.level
            try:
                self.LOG.root.level=logging.DEBUG
                self.LOG.error(format_text(msg, 'error'))
            finally:
                self.LOG.root.level=saveLevel

    def warn(self, msg, force=False):
        if force or self.level >= log_levels['warning']: 
            saveLevel = self.LOG.root.level
            try:
                self.LOG.root.level=logging.DEBUG
                self.LOG.warning(format_text(msg, 'warning'))
            finally:
                self.LOG.root.level=saveLevel

    def info(self, msg, stress=True, force=False):
        if force or self.level >= log_levels['info']: 
            saveLevel = self.LOG.root.level
            try:
                self.LOG.root.level=logging.DEBUG
                self.LOG.info(format_text(msg, 'info'))
            finally:
                self.LOG.root.level=saveLevel

    

    ### log debug message if (requested) log_level is greater than LOG_LEVEL_debug
    def debug(self, msg, force=False):
        if force or self.level >= log_levels['debug']: 
            saveLevel = self.LOG.root.level
            try:
                self.LOG.root.level=logging.DEBUG
                self.LOG.debug(format_text(msg, 'debug'))
            finally:
                self.LOG.root.level=saveLevel

    

    # ### always log
    # def force(self, msg, stress=True):
    #     saveLevel = self.LOG.root.level
    #     try:
    #         self.LOG.root.level=logging.DEBUG
    #         self.LOG.info(format_text(msg, 'bold' if stress else 'normal'))
    #     finally:
    #         self.LOG.root.level=saveLevel





# boto3.set_stream_logger('boto3.resources', logging.ERROR)

Logger = LoggerClass(log_levels[DEFAULT_LOG_LEVEL])




CLI_COMMANDS_HELP = {
    'parameter': {
        'add': """Add a parameter to a context, or update a parameter if it is already existing in the context.\n
                ** Tips: 1) If the parameter is inherited from the parent contexts, it will be overriden in the context. 2) Multiple parameters may be added at once using the '--input' option.\n
                KEY: The identifier of the parameter to be added. It may also be provided using the '--key' option.\n
                VALUE: The value for the parameter. If the parameter is already existing in the context, the value will be updated to the new one. It may also be provided using the '--value' option.\n
                """,
        'list': """Return the list of owned/inherited parameters in a context.\n
                ** Tips: 1) To limit the list to the owned/overriden parameters only, use the '--uninherited' option. This will return only the context specific parameters.
                """,
        'get': """Return the current value of a parameter.\n
                ** Tips: 1) The parameter may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the parameter. It may also be provided using the '--key' option.\n
                """,
        'edit': """Edit the current value of a parameter.\n
                ** Tips: 1) The parameter may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the parameter. It may also be provided using the '--key' option.\n
                """,
        'delete': """Delete a parameter from a context.\n
                ** Tips: 1) The inherited parameters cannot be deleted. The context must be the owner of the parameter or a not found error will be returned. 2) Multiple parameters may be deleted at once using the '--input' option.\n         
                KEY: The identifier of the parameter to be deleted. It may also be provided using the '--key' option.\n
                """,
        'history': """Return the revision history of parameter.\n
                ** Tips: 1) The parameter may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the parameter. It may also be provided using the '--key' option.\n
                """,
    },
    'secret': {
        'add': """Add a secret to a context, or update a secret if it is already existing in the context.\n
                ** Tips: 1) If the secret is inherited from the parent contexts, it will be overriden in the context. 2) Multiple secrets may be added at once using the '--input' option.\n
                KEY: The identifier of the secret to be added. It may also be provided using the '--key' option.\n
                VALUE: The value for the secret. If the secret is already existing in the context, the value will be updated to the new one. It may also be provided using the '--value' option.\n
                """,
        'list': """Return the list of owned/inherited secrets in a context.\n
                ** Tips: 1) To limit the list to the owned/overriden secrets only, use the '--uninherited' option. This will return only the context specific secrets.
                """,
        'get': """Return the current value of a secret.\n
                ** Tips: 1) The secret may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the secret. It may also be provided using the '--key' option.\n
                """,
        'edit': """Edit the current value of a secret.\n
                ** Tips: 1) The secret may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the secret. It may also be provided using the '--key' option.\n
                """,
        'delete': """Delete a secret from a context.\n
                ** Tips: 1) The inherited secrets cannot be deleted. The context must be the owner of the secret or a not found error will be returned. 2) Multiple secrets may be deleted at once using the '--input' option.\n         
                KEY: The identifier of the secret to be deleted. It may also be provided using the '--key' option.\n
                """,
        'history': """Return the revision history of secret.\n
                ** Tips: 1) The secret may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the secret. It may also be provided using the '--key' option.\n
                """,

    },
    'template': {
        'add': """Add a template to a context, or update the contents if it is already existing in the context.\n
                ** Tips: 1) If the template is inherited from the parent contexts, it will be overriden in the context. 2) Multiple templates may be added recursively from a directory.\n
                KEY: The identifier of the template to be added. It may also be provided using the '--key' option.\n
                """,
        'list': """Return the list of templates in a context.\n
                ** Tips: 1) To limit the list to the owned/overriden templates only, use the '--uninherited' option. This will return only the context specific templates.
                """,
        'get': """Return the contents of a template.\n
                ** Tips: 1) The template may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the secret. It may also be provided using the '--key' option.\n
                """,
        'edit': """Edit the contents of a template.\n
                ** Tips: 1) The template may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the secret. It may also be provided using the '--key' option.\n
                """,
        'delete': """Delete a template from a context.\n
                ** Tips: 1) The inherited template cannot be deleted. The context must be the owner of the secret or a not found error will be returned. 2) Multiple templates may be deleted at once using the '--input' option.\n
                KEY: The identifier of the template to be deleted. It may also be provided using the '--key' option.\n
                """,
        'history': """Return the revision history of template.\n
                ** Tips: 1) The template may be inherited from the parent contexts or owned by the given context.\n
                KEY: The identifier of the template. It may also be provided using the '--key' option.\n
                """,
        'render': """Render templates in a context.\n
                    """,
        },
    'config': {
        'get': """Get DSO application configuration.\n
                ** Tips: 1) Use --local or --global to get local or global configuration only.\n
                KEY: The key of the configuration
                """,
        'set': """Set DSO application configuration.\n
                ** Tips: 1) Use --local or --global to get local or global configuration only.\n
                KEY: The key of the configuration. It may also be provided using the '--key' option.\n
                VALUE: The value for the configuration. It may also be provided using the '--value' option.\n
                """,
        'unset': """Unset DSO application configuration.\n
                ** Tips: 1) Use --local or --global to get local or global configuration only.\n
                KEY: The key of the configuration
                """,
        'init': """Initialize DSO configuration for the working directory.\n
                ** Tips: 1) Use --input to load connfiguration from a file.\n
                The option '--working-dir' can be used to specify a different working directory than the current directory where dso is running in.\n
                """,
    },
    'network': {
        'layout_subnet_plan': """Layout subnet plan.\n
                """,

    }
}

CLI_COMMANDS_SHORT_HELP = {
    'version': "Display versions.",
    'parameter': {
        'list': "List parameters available to the application.",
        'add': "Add/Update one or multiple parameters to the application.",
        'get': "Get the value of a parameter.",
        'edit': "Edit the value of a parameter.",
        'delete': "Delete one or multiple parameters from the application.",
        'history': "Get the revision history of a parameter.",
    },
    'secret': {
        'list': "List secrets available to the application.",
        'add': "Add/Update one or multiple secrets to the application.",
        'get': "Get the value of a secret.",
        'edit': "Edit the value of a secret.",
        'delete': "Delete one or multiple secrets from the application.",
        'history': "Get the revision history of a secret.",
    },
    'template': {
        'list': "List templates available to the application.",
        'add': "Add/Update a template to the application.",
        'get': "Get the contents of a template.",
        'edit': "Edit the contents of a template.",
        'delete': "Delete one or multiple templates from the application.",
        'history': "Get the revision history of a template.",
        'render': "Render templates using parameters in a context.",
    },
    'package': {
        'list': "List packages built for the application.",
        'create': "Create a build package for the application.",
        'get': "Download an application build package.",
        'delete': "Delete a build package from the application.",
    },
    'release': {
        'list': "List deployment releases for the application.",
        'create': "Create a deployment release for the application.",
        'get': "Download an application deployment release.",
        'delete': "Delete a deployment release from the application.",
    },
    'config': {
        'get': "Get DSO application configuration(s).",
        'set': "Set the DSO application configuration(s).",
        'unset': "Delete a DSO application configuration.",
        'init': "Initialize DSO configuration for the working directory.",
    },
    'network': {
        'layout_subnet_plan': "Layout subnet plan group.",

    }

}
CLI_PARAMETERS_HELP = {
    'common': {
        'working_dir': "Path to a (local) directory where the DSO application configuration resides. By default, the current working directory will be used if the option is not provided.",
        'verbosity' : "Specify the logging verbosity, where 0 is for logging critical fatal errors only, 1 also logs error messages, 2 also logs warnings, 3 also logs information messages, 4 also logs debug messages, and finally 5 logs everything.",
        'stage' : "Requested a specific stage using the stage identifier, which is combination of a name and an optional number, where name must conform to ^([a-zA-Z][a-zA-Z0-9]+)$ regex expression. If no /<number> is specified, the default environment (/0) in the given context will be requested.",
        'input' : "Path to a (local) file defining the input data. Use '-' to read from the shell pipe or stdin. Use '--format' to specify the format if needed.",
        'format': "Specify the format of the output or the input if mixed with the '--input' option.",
        'config': "Comma separated list of key/value pairs to temporarily override the current DSO application configurations. It takes effect only while executing the command and does not have any lasting effect on the DSO application configuration or subsequent command executions.",
        'query': "Customize output using JMESPath query language.",
        'query_all': "Include all the available fields in the ouput.",
        'global_scope': "Use the global scope.",
        'project_scope': "Use the project scope.",
        'scope': "Select the context scope.",
        'filter': "Use a regex pattern to filter result by the provider.",
        'context': "Specify the context to be used.",
        'namespace': "Override the namespace in the context.",
        'project': "Override the project in the context.",
        'application': "Override the application in the context.",

    },
    'parameter': {
        'key': "The key of the parameter. See KEY argument for more details.",
        'value': "The value for the parameter. See VALUE argument for more details.",
        'query_values': "Include parameter values in the output.",
        'uninherited': "Select only parameters which are specific to the gievn context, i.e. not inherited from the parent contexts.",
        'revision': "The revision ID whose value to be fetched.",
        'history': "Get the revision history of the parameter.",

    },
    'secret': {
        'key': "The key of the secret",
        'value': "The value for the secret",
        'decrypt': "Decrypt the secret value.",
        'query_values': "Include secret values in the output.",
        'uninherited': "Select only secrets which are specific to the gievn context, i.e. not inherited from the parent contexts.",
        'revision': "The revision ID whose value to be fetched.",
        'history': "Get the revision history of the secret.",
    },
    'template': {
        'type': "Type of the template. Use 'resource' for templates needed at the provision time when provisioning resources required by the application to run such as SQS queus, SNS topics, and CI/CD piplines.\nUse 'package' for templates needed at the build time when generating a package.\nUse 'release' for templates needed at the deploy time when generating a release." ,
        'key': "The key of the template",
        'limit': "Limit templates to be rendered.",
        'render_path': "Path (relative to the root of the DSO application) where rendered template will be placed at.",
        'query_render_path': "Include the template render paths in the output.",
        'contents_path' : "Path to a local file or directory containing the template contents.",
        'recursive' : "Add files recursively.",
        'uninherited': "Select only templates which are specific to the gievn context, i.e. not inherited from the parent contexts.",
        'include_contents': "Include template contenets in the output. ",
        'history': "Get the revision history of the template.",

    },
    'config': {
        'key': "The key of the configuration",
        'value': 'Value for the configuration key',
        'input' : "Path to a local (yaml) file inputing the configuration. Use '-' to read from the shell pipe or stdin.",
        'local': "Select the local DSO configurations, i.e. existing in the working directory.",
        'global': "Select the global DSO configurations, i.e. user-wide configuration.",
        'override_inherited': "Explicitly override inherited configurations locally.",
        'setup': "Run a setup wizard to assist configuring the DSO application.",

    },
    'network': {
        'subnet_layout_mode': "Select the subnet plan layout mode.",
    
    }


}


CLI_MESSAGES = {
    'MissingOption': "Missing option {0}.",
    'MissingArgument': "Missing argument {0}.",
    'ArgumentsProvided': "The following argument/option must be provided: {1}",
    'ArgumentsOnlyOneProvided': "Only one of the following arguments/options may be provided: {0}",
    'ArgumentsAtLeastOneProvided': "At least one of the following arguments/options must be provided: {0}",
    'ArgumentsAllProvided': "All of the following arguments/options must be provided: {0}",
    'ArgumentsNoneProvided': "The following arguments/options cannot be provided: {0}",
    'ArgumentsNotAllProvided': "The following arguments/options cannot be provided together: {0}",
    'ArgumentsProvidedBecause': "Since {0} provided, the following argument/option must also be provide: {1}",
    'ArgumentsOnlyOneProvidedBecause': "Since {0} provided, only one of the following arguments/options may be provided: {1}",
    'ArgumentsAtLeastOneProvidedBecause': "Since {0} provided, at least one of the following arguments/options must also be provided: {1}",
    'ArgumentsAllProvidedBecause': "Since {0} provided, all of the following arguments/options must also be provide: {1}",
    'ArgumentsNoneProvidedBecause': "Since {0} provided, the following arguments/options cannot be provided: {1}",
    'ArgumentsNotAllProvidedBecasue': "Since {0} provided, the following arguments/options may be provided together: {1}",
    'TryHelpWithCommand': "Try '{0} --help' for more details.",
    'TryHelp': "Try the command with '-h' / '--help' option for more details.",
    'InvalidFileFormat': "Invalid file, not conforming to the expected '{0}' format.",
    'EnteredSecretValuesNotMatched': "Entered values for the secret did not macth.",
    'RenderPathNotReleative': "Render path '{0}' is not releative to the application root directory.",
    # 'InvalidRenderPath': "'{0}' is not a valid render path.",
    'InvalidRenderPathExistingDir': "'{0}' is not a valid render path because it is an existing directory.",
    'InvalidFilter': "Invalid regex pattern for filter: {0}",
    'NoChanegeDetectedAfterEditing': "No change detected after editing.",
    'ParameterNotFound': "Parameter '{0}' not found in the given context: namespace={1}, project={2}, application={3}, stage={4}, scope={5}",
    'SecretNotFound': "Secret '{0}' not found in the given context: namespace={1}, project={2}, application={3}, stage={4}, scope={5}",
    'TemplateNotFound': "Template '{0}' not found in the given context: namespace={1}, project={2}, application={3}, stage={4}, scope={5}",
    'MissingField': "Missing field: {0}",
}



import sys
import os
import ipaddress
import re
from .logger import Logger
from .exceptions import DSOException
from .dict_utils import *

UNNAMED_SUBNET = 'N/A'

def ipv4_to_binary(ip, sep="."):
    return  sep.join(map(str,["{0:08b}".format(int(x)) for x in ip.split(".")]))


class IPv4AddressX(ipaddress.IPv4Address):
    def to_binary(self, sep="."):
        """Represent IPv4 as 4 blocks of 8 bits."""
        return sep.join(f"{i:08b}" for i in self.packed)

    @classmethod
    def from_binary_repr(cls, binary_repr: str):
        """Construct IPv4 from binary representation."""
        # Remove anything that's not a 0 or 1
        i = int(re.sub(r"[^01]", "", binary_repr), 2)
        return cls(i)



def layout_subnet_plan(subnet_plan, filters=None, summary=True):

    def add_segment_subnets(segments, segment_ix, supernet, result: dict):

        def is_subnet_selected(supernet, subnet, selects):
            if not subnet.subnet_of(supernet):
                return False
            subnetIP = IPv4AddressX(subnet.network_address).to_binary(sep='')
            result = True
            for select in map(lambda x: x.strip(), selects.split('|')):
                try:
                    invert, operation, base, msd_dontcare, digits, lsd_dontcare = re.match('^(!|not)?\s*(msd|lsd)\(0(\w)([x]*)([01]*)([x]*)\)$', select).groups()[0:]
                except:
                    raise DSOException(f"Invalid value for 'select': {select}")
                invert = 'not' if invert == '!' else '' if invert is None else invert
                if not base in ['b']:
                    raise NotImplementedError
                ### most significant digit
                if operation == 'msd':
                    target = subnetIP[supernet.prefixlen + len(msd_dontcare):]
                ### least significant digit
                elif operation == 'lsd':
                    target = subnetIP[supernet.max_prefixlen - len(digits) - len(lsd_dontcare):]
                selected = eval(f"{invert} target.startswith(digits)")
                result = result and selected
            return result

        ### if last segment, add host count
        if segment_ix == len(segments):
            hosts = list(supernet.hosts())
            result['Hosts'] = {
                'Count': len(hosts),
                'IP': f"{hosts[0]}/32-{hosts[-1]}/32"
            }
            return

        segment = segments[segment_ix]
        # key = segment.get('plural', f"{segment.get('name')}s")
        key = segment.get('name')
        if segment['bits'] > 0:
            subnets = list(supernet.subnets(segment['bits']))
            result[key] = {
                'Count': 0, ### to be incremented while adding subnets
                'Ranges': []
            }
            range_first_subnet = None
            range_last_subnet = None
            for i, subnet in enumerate(subnets):
                if 'select' in segment:
                    if not is_subnet_selected(supernet, subnet, segment['select']): 
                        ### add the last subnet in the range if in summary mode
                        if summary and not range_last_subnet is None:
                            result[key]['Ranges'][-1]['Subnets'].append({
                                'CIDR' : range_last_subnet.with_prefixlen,
                                'Number': result[key]['Count'] + 1,
                                'Name' : segment['labels'].get(str(result[key]['Count']-1), UNNAMED_SUBNET),
                            })
                            add_segment_subnets(segments, segment_ix+1, range_last_subnet, result[key]['Ranges'][-1]['Subnets'][-1])

                        range_first_subnet = None
                        range_last_subnet = None
                        Logger.debug(f"{key} {i} excluded: {subnet}")
                        continue

                ### if start of a range, append a new Range
                if range_first_subnet is None:
                    range_first_subnet = subnet
                    result[key]['Ranges'].append({
                        'CIDR': range_first_subnet.with_prefixlen,
                        'Count': 0, ### to be incremented while adding subnets
                        'Subnets': []
                    })
                    if summary:
                        result[key]['Ranges'][-1]['Subnets'].append({
                            'CIDR': range_first_subnet.with_prefixlen,
                            'Number': result[key]['Count'] + 1,
                            'Name': segment['labels'].get(result[key]['Count'], UNNAMED_SUBNET),
                        })
                        add_segment_subnets(segments, segment_ix+1, range_first_subnet, result[key]['Ranges'][-1]['Subnets'][-1])
                else:
                    range_last_subnet = subnet
                    cidr = f"{range_first_subnet.with_prefixlen}-{range_last_subnet.with_prefixlen}"
                    ### updare Range CIDR
                    result[key]['Ranges'][-1]['CIDR'] = cidr
                    ### if in summary mode, always add the last subnet in the list
                    if summary and range_last_subnet == subnets[-1]:
                        result[key]['Ranges'][-1]['Subnets'].append({
                            'CIDR' : range_last_subnet.with_prefixlen,
                            'Number': result[key]['Count'] + 1,
                            'Name': segment['labels'].get(result[key]['Count'], UNNAMED_SUBNET),
                        })
                        add_segment_subnets(segments, segment_ix+1, range_last_subnet, result[key]['Ranges'][-1]['Subnets'][-1])

                ### always add the subnet if not in summary mode
                if not summary:
                    cidr = {
                            'CIDR' : subnet.with_prefixlen,
                            'Number': result[key]['Count'] + 1,
                            'Name': segment['labels'].get(result[key]['Count'], UNNAMED_SUBNET),
                    }
                    selected = True
                    if filters and 'selector' in filters:
                        if key in filters['selector']:
                            for k,v in filters['selector'][key].items():
                                selected = selected and k in cidr and cidr[k] == v
                                if not selected: break
                    if selected:
                        result[key]['Ranges'][-1]['Subnets'].append(cidr)
                        add_segment_subnets(segments, segment_ix+1, subnet, result[key]['Ranges'][-1]['Subnets'][-1])

                result[key]['Count'] = result[key]['Count'] + 1
                result[key]['Ranges'][-1]['Count'] = result[key]['Ranges'][-1]['Count'] + 1

            # ### comporess Range list if has only one element
            # if compress and len(result[key]['Ranges']) == 1:
            #     result[key].update(result[key]['Ranges'][0])
            #     result[key].pop('Ranges')

        ### skip segments with 0 bits long, and move to the next segment
        else:
            add_segment_subnets(segments, segment_ix+1, supernet, result)


    ####################################################################################

    if filters:
        summary = False
        if 'plan' in filters and filters['plan']:
            plans = list(filter(lambda x: x['name'] == filters['plan'], subnet_plan['plans']))
            if not plans:
                raise DSOException(f"Plan '{filters['plan']}' not found in the given subnet plan.")
    else:
        plans = subnet_plan['plans']

    plans = list(filter(lambda x: str(x['enabled']).lower() in ['1', 'yes', 'true'], plans))
    if not plans:
        if filters['plan']:
            raise DSOException(f"Plan '{filters['plan']}' is disabled in the given subnet plan.")
        else:
            raise DSOException(f"No active plan found in the given subnet plan.")

    layouts = []
    for plan in plans:
        layouts.append({})
        add_segment_subnets(plan['segments'], 0, ipaddress.ip_network(subnet_plan['cidr']), layouts[-1])


    merged = {}
    for layout in layouts:
        merge_dicts(layout, merged, merge_key='CIDR')

    result = {'SubnetPlanLayout': {
                'Name': subnet_plan['name'],
                'Description': subnet_plan['description'],
                'CIDR': subnet_plan['cidr'],
                'Layout': merged
            }
    }

    return result





import os
from .logger import Logger
from .exceptions import DSOException

def clean_directory(path):
    for path in Path(path).glob("**/*"):
        if path.is_file():
            path.unlink()
        elif path.is_dir():
            rmtree(path)




def is_binary_file(filename):
    """ 
    Return true if the given filename appears to be binary.
    File is considered to be binary if it contains a NULL byte.
    FIXME: This approach incorrectly reports UTF-16 as binary.
    """
    with open(filename, 'rb') as f:
        for block in f:
            if b'\0' in block:
                return True
    return False


def exists_on_path(filename):
    return any([os.path.exists(os.path.join(p, filename)) for p in os.environ.get('PATH', '').split(os.pathsep)])


def render_stream(stream, values):
    if not values: return stream
    import jinja2
    template = jinja2.Environment(undefined=jinja2.StrictUndefined).from_string(stream)
    try:
        rendered = template.render(values)
    except Exception as e:
        Logger.error(f"Failed to render stream.")
        msg = getattr(e, 'message', getattr(e, 'msg', str(e)))
        raise DSOException(msg)

    return rendered


def get_format_from_file_name(file_name):
    from os.path import splitext
    ext = splitext(file_name)[1]
    if ext in ['.yml', '.yaml']:
        return 'yaml'
    elif ext in ['.json', '.csv']:
        return ext[1:]
    else:
        return 'raw'


def load_file(file_path, format='auto', pre_render_values=None):
    if is_binary_file(file_path):
        raise DSOException(f"Cannot load binary file '{file_path}'.")
    
    if format == 'auto':
        format = get_format_from_file_name(file_path)
    
    with open(file_path, 'r', encoding='utf-8') as f:
        if format == 'raw':
            return render_stream(f.read(), pre_render_values)
        elif format == 'yaml':
            import yaml
            return yaml.safe_load(render_stream(f.read(), pre_render_values)) or {}
        elif format == 'json':
            import json
            return json.loads(render_stream(f.read(), pre_render_values) or '{}') or {}
        elif format == 'csv':
            import csv
            return list(csv.reader(render_stream(f.read(), pre_render_values))) or []
        else:
            raise NotImplementedError



def save_data(data, file_path, format='auto'):
    if format == 'auto':
        format = get_format_from_file_name(file_path)

    import os
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, 'w') as f:
        if format == 'yaml':
            import yaml
            yaml.dump(data, f, sort_keys=False, indent=2)
        elif format == 'json':
            import json
            json.dump(data, f, sort_keys=False, indent=2)
        else:
            raise NotImplementedError


def get_file_modified_date(file_path, format='%A, %Y-%m-%d %H:%M:%S'):
    from time import strftime, localtime
    return strftime(format, localtime(os.path.getmtime(file_path)))


import sys
import os
import click
from .logger import Logger
from .cli_constants import *
import tempfile

class MuOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if: list = kwargs.pop("not_required_if")

        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs["help"] = (kwargs.get("help", "") + " This option is mutually exclusive with --" + ", --".join(self.not_required_if) + ".").strip()
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt: bool = self.consume_value(ctx, opts)
        for other_param in ctx.command.get_params(ctx):
            if other_param is self:
                continue
            if other_param.human_readable_name in self.not_required_if:
                other_opt: bool = other_param.consume_value(ctx, opts)
                if other_opt:
                    if current_opt:
                        Logger.error(CLI_MESSAGES['ArgumentsMutalExclusive'].format(str(self.name), str(other_param.human_readable_name)))                      
                        Logger.info(CLI_MESSAGES['TryHelpWithCommand'].format(ctx.command_path), stress=False)
                        ctx.abort()
                    else:
                        self.required = None
        return super().handle_parse_result(ctx, opts, args)


# def modify_click_usage_error(main_command):
def modify_click_usage_error():

    def show(self, file=None):
        import sys
        # if file is None:
        #     file = get_text_stderr()
        Logger.error(self.format_message())
        if self.ctx:
            # echo(self.ctx.get_usage() + '\n', file=file, color=color)
            # echo(self.ctx.get_help(), color=self.ctx.color)
            # echo(CLI_MESSAGES['TryHelp'].format(self.ctx.command_path), color=self.ctx.color)
            # Logger.info(CLI_MESSAGES['TryHelpWithCommand'].format(self.ctx.command_path), force=True)
            Logger.info(CLI_MESSAGES['TryHelpWithCommand'].format(self.ctx.command_path), stress=False, force=True)
        else:
            ### for some reason self.ctx is None for BadOptionUsage exception, maybe a bug in click package,
            ### hence can't get the command_path
            # for i in dir(self):
            #     print(f"{i}=", getattr(self, i))
            Logger.info(CLI_MESSAGES['TryHelp'], stress=False, force=True)

        sys.argv = [sys.argv[0]]
        # main_command()

    click.exceptions.UsageError.show = show


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))
        
class CustomMultiGroup(click.Group):

    def group(self, *args, **kwargs):
        """Behaves the same as `click.Group.group()` except if passed
        a list of names, all after the first will be aliases for the first.
        """
        def decorator(f):
            aliased_group = []
            if isinstance(args[0], list):
                # we have a list so create group aliases
                _args = [args[0][0]] + list(args[1:])
                for alias in args[0][1:]:
                    grp = super(CustomMultiGroup, self).group(
                        alias, *args[1:], **kwargs)(f)
                    grp.short_help = "Alias for '{}'".format(_args[0])
                    aliased_group.append(grp)
            else:
                _args = args

            # create the main group
            grp = super(CustomMultiGroup, self).group(*_args, **kwargs)(f)

            # for all of the aliased groups, share the main group commands
            for aliased in aliased_group:
                aliased.commands = grp.commands

            return grp

        return decorator


def command_doc(value):
    def _doc(func):
        func.__doc__ = value
        return func
    return _doc


def check_file_path(ctx, self, value):
    if not value: return
    if value == '-':
        with tempfile.NamedTemporaryFile("w", delete=False) as f:
            f.write(sys.stdin.read())
            value = f.name
    else:
        if not os.path.exists(value):
            raise click.BadParameter(f"'{value}' does not exist.")
    return value


import re
from .logger import Logger
from .providers import StoreProvider, Providers
from .stages import Stages
from .constants import *
from .exceptions import *
from .config import ContextMode


REGEX_PATTERN  = r"^[a-zA-Z]([.a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class ParameterProvider(StoreProvider):
    def list(self, config, uninherited=False, filter=None):
        raise NotImplementedError()
    def add(self, config, key, value):
        raise NotImplementedError()
    def get(self, config, key, revision=None):
        raise NotImplementedError()
    def history(self, config, key):
        raise NotImplementedError()
    def delete(self, config, key):
        raise NotImplementedError()


class ParameterService():

    def validate_key(self, key):
        Logger.info(f"Validating parameter key '{key}'...")
        if not key:
            raise DSOException(MESSAGES['KeyNull'])
        if key == 'dso' or key.startswith('dso.'):
            raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
        if not re.match(REGEX_PATTERN, key):
            raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, REGEX_PATTERN))
        if '..' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))
            
    def list(self, config, uninherited=False, filter=None):
        self.config = config
        provider = Providers.ParameterProvider()
        Logger.info(f"Start listing parameters: namespace={config.get_namespace(ContextMode.Requested)}, project={config.get_project(ContextMode.Requested)}, application={config.get_application(ContextMode.Requested)}, stage={config.get_short_stage(ContextMode.Requested)}, scope={config.context.scope}")
        return provider.list(config, uninherited, filter)

    def add(self, config, key, value):
        self.config = config
        self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Start adding parameter '{key}': namespace={config.get_namespace(ContextMode.Requested)}, project={config.get_project(ContextMode.Requested)}, application={config.get_application(ContextMode.Requested)}, stage={config.get_short_stage(ContextMode.Requested)}, scope={config.context.scope}")
        return provider.add(config, key, value)

    def get(self, config, key, revision=None):
        self.config = config
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Start getting parameter '{key}': namespace={config.get_namespace(ContextMode.Requested)}, project={config.get_project(ContextMode.Requested)}, application={config.get_application(ContextMode.Requested)}, stage={config.get_short_stage(ContextMode.Requested)}, scope={config.context.scope}")
        return provider.get(config, key, revision)

    def history(self, config, key):
        self.config = config
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Start getting the history of parameter '{key}': namespace={config.get_namespace(ContextMode.Requested)}, project={config.get_project(ContextMode.Requested)}, application={config.get_application(ContextMode.Requested)}, stage={config.get_short_stage(ContextMode.Requested)}, scope={config.context.scope}")
        return provider.history(config, key)

    def delete(self, config, key):
        self.config = config
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Start deleting parameter '{key}': namespace={config.get_namespace(ContextMode.Requested)}, project={config.get_project(ContextMode.Requested)}, application={config.get_application(ContextMode.Requested)}, stage={config.get_short_stage(ContextMode.Requested)}, scope={config.context.scope}")
        return provider.delete(config, key)

Parameters = ParameterService()



import re
import os
import sys
import imp
import yaml
from .enum_utils import OrderedEnum
from .constants import *
from .logger import Logger
from .dict_utils import *
from pathlib import Path
from .file_utils import *
from .exceptions import DSOException
from .stages import Stages
from .constants import *
from .exceptions import DSOException


class ContextScope(OrderedEnum):
    Application = 10
    Project = 20
    Global = 30


class ConfigScope(OrderedEnum):
    Local = 10
    Middle = 20
    Global = 30


context_translation_matrix = {
    'default': {
        'default': {
            'default': {
                'default': {
                    '0': "Global",
                },
                'stage': {
                    '0': "Global Stage",
                    'n': "Global Numbered Stage",
                },
            },
        },
        'project': {
            'default': {
                'default': {
                    '0': "Project",
                },
                'stage': {
                    '0': "Project Stage",
                    'n': "Project Numbered Stage",
                },
            },
            'application': {
                'default': {
                    '0': "Application",
                },
                'stage': {
                    '0': "Application Stage",
                    'n': "Application Numbered Stage",
                },
            },
        },
    },
    'namespace': {
        'default': {
            'default': {
                'default': {
                    '0': "Global",
                },
                'stage': {
                    '0': "Global Stage",
                    'n': "Global Numbered Stage",
                },
            },
        },
        'project': {
            'default': {
                'default': {
                    '0': "Project",
                },
                'stage': {
                    '0': "Project Stage",
                    'n': "Project Numbered Stage",
                },
            },
            'application': {
                'default': {
                    '0': "Application",
                },
                'stage': {
                    '0': "Application Stage",
                    'n': "Application Numbered Stage",
                },
            },
        },
    }

}


_default_config = {
    'kind': 'dso/config',
    'version': 1,
    # 'namespace': 'default',
    # 'project': 'default',
    # 'application': 'default',
    # 'stage': 'default',
    'contexts': [
        {
            'name': 'default',
            'spec': {
                'namespace': 'default',
                'project': 'default',
                'application': 'default',
                'stage': 'default',
            }
        },
    ],
}


def get_default_config():
    return _default_config.copy()



class Context():

    def __init__(self, namespace=None, project=None, application=None, stage=None, scope=ContextScope.Application):
        # self.namespace = namespace
        # self.project = project
        # self.application = application
        # self.stage = stage
        # self._scope = scope
        
        self.set_namespace(namespace)
        self.set_project(project)
        self.set_application(application)
        self.set_stage(stage)
        self._scope = scope

    def __str__(self):
        return f"namespace={self.namespace}, project={self.project}, application={self.application}, stage={self.short_stage}"


    @property
    def namespace(self):
        return self._namespace

    def set_namespace(self, value):
        if value:
            self._namespace = value
        else:
            if 'DSO_NAMESPACE' in os.environ:
                self._namespace = os.getenv('DSO_NAMESPACE').lower() or 'default'
            else:
                self._namespace = 'default'


    @property
    def project(self):
        return self.get_project()

    def get_project(self, ignore_scope=False):
        if ignore_scope:
            result = self._project
        else:
            if self._scope > ContextScope.Project:
                result = 'default'
                Logger.warn("Switched to the global scope.")
            else:
                result = self._project
        return result

    def set_project(self, value):
        if value:
            self._project = value
        else:
            if 'DSO_PROJECT' in os.environ:
                self._project = os.getenv('DSO_PROJECT').lower() or 'default'
            else:
                self._project = 'default'

    @property
    def application(self):
        return self.get_application()

    def get_application(self, ignore_scope=False):
        if ignore_scope:
            result = self._application
        else:
            if self._scope > ContextScope.Application:
                result = 'default'
                Logger.warn("Switched to the project scope.")
            else:
                result = self._application
        return result


    def set_application(self, value):
        if value:
            self._application = value
        else:
            if 'DSO_APPLICATON' in os.environ:
                self._application = os.getenv('DSO_APPLICATON').lower() or 'default'
            else:
                self._application = 'default'

    @property
    def stage(self):
        return self._stage

    def set_stage(self, value):
        if value:
            self._stage = value
        else:
            if 'DSO_STAGE' in os.environ:
                self._stage = os.getenv('DSO_STAGE').lower() or Stages.default_stage
            else:
                self._stage = Stages.default_stage
        
        self._stage = Stages.normalize( self._stage)
        self._short_stage = Stages.shorten(self._stage)


    @property
    def short_stage(self):
        return self._short_stage


    @property
    def scope(self):
        return self._scope


    @property
    def scope_translation(self):
        namespace_idx = 'default' if self.namespace == 'default' else 'namespace'
        project_idx = 'default' if self.project == 'default' else 'project'
        application_idx = 'default' if self.application == 'default' else 'application'
        stage_idx = 'default' if Stages.is_default(self.stage) else 'stage'
        n_idx = '0' if Stages.is_default_env(self.stage) else 'n'
        return context_translation_matrix[namespace_idx][project_idx][application_idx][stage_idx][n_idx]


    def get_path(self, key=None):
        result = f"/{self.namespace}/{self.project}"
        ### every application must belong to a project, no application overrides allowed in the default project
        if not self.project == 'default':
            result += f"/{self.application}"
        else:
            result += "/default"
        result += f"/{self.stage}"
        if key:
            result += f"/{key}"
        return result


class ContextService():

    @property
    def config_dir(self):
        return '.dso'

    @property
    def config_file(self):
        return 'config.yml'

    context = Context()
    working_dir = ''
    local_config = {}
    local_config_file_path = ''
    local_config_dir_path = ''
    global_config = {}
    global_config_file_path = ''
    global_config_dir_path = ''
    inherited_config = {}
    inherited_config_files = []
    merged_config = {}


    def load(self, working_dir, context_name=None, namespace=None, project=None, application=None, stage=None, scope=ContextScope.Application):
        self.working_dir = working_dir
        self.update_merged_config()
        self.load_global_config()
        self.load_inherited_config()
        self.load_local_config()
        self.load_context(context_name, namespace, project, application, stage, scope)
        self.check_version()


    @property
    def meta_vars(self):
        return {'dso': self.merged_config}


    def load_context(self, context_name, namespace, project, application, stage, scope):

        if context_name:
            found = self.get_contexts(context_name)
            if not found:
                raise DSOException(f"Context '{context_name}' not found.")
            self._namespace = found[-1]['spec']['namespace']
            self._project = found[-1]['spec']['project']
            self._application = found[-1]['spec']['application']
            self._stage = found[-1]['spec']['stage']
        else:
            self._namespace = 'default'
            self._project = 'default'
            self._application = 'default'
            self._stage = 'default'

        if namespace:
            if not namespace == self._namespace:
                self._namespace = namespace
                Logger.debug(f"Namespace overidden from '{self._namespace}' to '{namespace}'.")
        else:
            if 'DSO_NAMESPACE' in os.environ:
                ev_namespace = os.getenv('DSO_NAMESPACE').lower() or 'default'
                if not ev_namespace == self._namespace:
                    self._namespace = ev_namespace
                    Logger.debug(f"Environment variable 'DSO_NAMESPACE' used.")

        if project:
            if not project == self._project:
                self._project = project
                Logger.debug(f"Project overidden from '{self._project}' to '{project}'.")
        else:
            if 'DSO_PROJECT' in os.environ:
                ev_project = os.getenv('DSO_PROJECT').lower() or 'default'
                if not ev_project == self._project:
                    self._project = ev_project
                    Logger.debug(f"Environment variable 'DSO_PROJECT' used.")

        if application:
            if not application == self._application:
                self._application = application
                Logger.debug(f"Project overidden from '{self._application}' to '{application}'.")
        else:
            if 'DSO_APPLICATION' in os.environ:
                ev_application = os.getenv('DSO_APPLICATION').lower() or 'default'
                if not ev_application == self._application:
                    self._application = ev_application
                    Logger.debug(f"Environment variable 'DSO_APPLICATION' used.")

        if stage:
            if not stage == self._stage:
                self._stage = stage
                Logger.debug(f"Project overidden from '{self._stage}' to '{stage}'.")
        else:
            if 'DSO_STAGE' in os.environ:
                ev_stage = os.getenv('DSO_STAGE').lower() or 'default'
                if not ev_stage == self._stage:
                    self._stage = ev_stage
                    Logger.debug(f"Environment variable 'DSO_STAGE' used.")
        self._short_stage = Stages.shorten(self._stage)

        self._scope = scope
        self.context = Context(self._namespace, self._project, self._application, self._stage, self._scope)


    def load_global_config(self):
        self.global_config = {}
        self.global_config_dir_path = os.path.join(Path.home(), self.config_dir)
        self.global_config_file_path = os.path.join(self.global_config_dir_path, self.config_file)

        if not os.path.exists(self.global_config_file_path):
            Logger.debug("Global DSO configuration not found.")
            return

        Logger.debug(f"Global DSO configuration found: path={self.global_config_file_path}")
        try:
            self.global_config_raw = load_file(self.global_config_file_path)
            self.global_config = load_file(self.global_config_file_path, pre_render_values=self.meta_vars)
        except:
            raise
            raise DSOException(MESSAGES['InvalidDSOConfigurationFile'].format(self.global_config_file_path))

        self.update_merged_config()


    def load_local_config(self, silent=False):
        self.local_config_raw = {}
        self.local_config = {}
        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)

        if not os.path.exists(self.local_config_file_path):
            if not silent:
                Logger.warn(MESSAGES['NoDSOConfigFound'])
            return

        Logger.debug(f"Local DSO configuration found: path={self.local_config_file_path}")

        try:
            self.local_config_raw = load_file(self.local_config_file_path)
            self.local_config = load_file(self.local_config_file_path, pre_render_values=self.meta_vars)
        except:
            raise
            raise DSOException(MESSAGES['InvalidDSOConfigurationFile'].format(self.local_config_file_path))

        self.update_merged_config()

    def load_inherited_config(self):

        self.inherited_config = {}

        for dir in Path(self.working_dir).resolve().parents:
            configFilePath = os.path.join(dir, self.config_dir, self.config_file)
            if os.path.exists(configFilePath):
                if not os.path.samefile(configFilePath, self.global_config_file_path):
                    Logger.debug(f"Inherited DSO configuration found: path={configFilePath}")
                    self.inherited_config_files.append(configFilePath)


        for configFilePath in reversed(self.inherited_config_files):
            try:
                config = load_file(configFilePath, self.meta_vars)
            except:
                raise
                raise DSOException(MESSAGES['InvalidDSOConfigurationFile'].format(configFilePath))
            
            self.inherited_config = merge_dicts(config, self.inherited_config)

        self.update_merged_config()



    def update_merged_config(self):
        self.merged_config = get_default_config()
        merge_dicts(self.global_config, self.merged_config)
        merge_dicts(self.inherited_config, self.merged_config)
        merge_dicts(self.local_config, self.merged_config)
        
        ### add context 
        self.merged_config['namespace'] = self.context.namespace
        self.merged_config['project'] = self.context.project
        self.merged_config['application'] = self.context.application
        self.merged_config['stage'] = self.context.stage
        self.merged_config['scope'] = str(self.context.scope)



    def check_version(self):
        if not int(self.merged_config['version']) == int(_default_config['version']):
            if int(self.merged_config['version']) > int(_default_config['version']):
                Logger.warn(MESSAGES['DSOConfigNewer'].format(_default_config['version'], self.merged_config['version']))
            else:
                Logger.warn(MESSAGES['DSOConfigOlder'].format(_default_config['version'], self.merged_config['version']))


    def save_local_config(self):
        os.makedirs(self.local_config_dir_path, exist_ok=True)
        save_data(self.local_config_raw, self.local_config_file_path)


    def save_global_config(self):
        os.makedirs(self.global_config_dir_path, exist_ok=True)
        save_data(self.global_config_raw, self.global_config_file_path)


    @property
    def contexts(self):
        return self.get_contexts()

    def get_contexts(self, name=None):
        try:
            result = self.merged_config['contexts']
        except nameError:
            raise DSOException("Invalid dso configuration schema.")
        if not name:
            return result
        else:
            found = [x for x in result if x['name'] == name]
            return found if found else []

    @property
    def namespace(self):
        return self.context.namespace

    # @property
    # def namespace(self):
    #     return self.get_namespace()
        
    # def get_namespace(self):
    #     if 'namespace' in self.local_config:
    #         result = self.local_config['namespace'].lower() or 'default'

    #     elif 'DSO_NAMESPACE' in os.environ:
    #         result = os.getenv('DSO_NAMESPACE').lower() or 'default'

    #     elif 'namespace' in self.inherited_config:
    #         result = self.inherited_config['namespace'].lower() or 'default'

    #     elif 'namespace' in self.global_config:
    #         result = self.global_config['namespace'].lower() or 'default'

    #     else:
    #         result = 'default'

    #     return result

    @property
    def project(self):
        return self.context.project


    # @property
    # def project(self):
    #     return self.get_project()

    # def get_project(self):
    #     if 'project' in self.local_config:
    #         result = self.local_config['project'].lower() or 'default'

    #     elif 'DSO_PROJECT' in os.environ:
    #         result = os.getenv('DSO_PROJECT').lower() or 'default'

    #     elif 'project' in self.inherited_config:
    #         result = self.inherited_config['project'].lower() or 'default'

    #     elif 'project' in self.global_config:
    #         result = self.global_config['project'].lower() or 'default'

    #     else:
    #         result = 'default'

    #     return result

    @property
    def application(self):
        return self.context.application


    # @property
    # def application(self):
    #     return self.get_application()

    # def get_application(self):
    #     if 'application' in self.local_config:
    #         result = self.local_config['application'].lower() or 'default'

    #     elif 'DSO_APPLICATION' in os.environ:
    #         result = os.getenv('DSO_APPLICATION').lower() or 'default'

    #     elif 'application' in self.inherited_config:
    #         result = self.inherited_config['application'].lower() or 'default'

    #     elif 'application' in self.global_config:
    #         result = self.global_config['application'].lower() or 'default'

    #     else:
    #         result = 'default'
        
    #     if not result == 'default' and self.project == 'default':
    #         Logger.debug(f"Application '{result}' was ignored because the global context is being used.")
    #         result = 'default'

    #     return result



    @property
    def stage(self):
        return self.context.stage



    @property
    def short_stage(self):
        return self.context.short_stage

    # @property
    # def stage(self):
    #     return self.get_stage()

    # def get_stage(self):
    #     if 'stage' in self.merged_config:
    #         return self.merged_config['stage']
    #     else: 
    #         return 'default'


    def get(self, key=None, scope=ConfigScope.Local):
        if key:
            Logger.info("Getting '{0}' from DSO configurations...".format(key))
        else:
            Logger.info("Getting DSO configurations...")

        if scope == ConfigScope.Local:
            usedConfig = merge_dicts(self.overriden_config, self.local_config.copy())
        elif scope == ConfigScope.Global:
            usedConfig = merge_dicts(self.overriden_config, self.global_config.copy())
        else:
            usedConfig = self.merged_config.copy()

        if key:
            result = get_dict_item(usedConfig, key.split('.'))
            if not result:
                raise DSOException(f"DSO configuration '{key}' not found.")
            return result
        else:
            return usedConfig

    def set(self, key, value, use_global):
        if use_global:
            Logger.info(f"Setting '{key}' to '{value}' in the global DSO configurations...")
            if not os.path.exists(self.local_config_file_path):
                raise DSOException("The global configuration has not been intitialized yet. Run 'dso config init --global' to initialize it.")

            set_dict_value(self.global_config_raw, key.split('.'), value, overwrite_parent=True, overwrite_children=True)
            self.save_global_config()
            self.load_global_config()
            # set_dict_value(self.global_config, key.split('.'), value, overwrite_parent=True, overwrite_children=True)
            # self.update_merged_config()
        else:
            Logger.info(f"Setting '{key}' to '{value}' in the local DSO configurations...")
            if not os.path.exists(self.local_config_file_path):
                raise DSOException("The working directory has not been intitialized yet. Run 'dso config init' to initialize it.")

            set_dict_value(self.local_config_raw, key.split('.'), value, overwrite_parent=True, overwrite_children=True)
            self.save_local_config()
            self.load_local_config()

    def delete(self, key, use_global):
        if use_global:
            Logger.info(f"Deleting '{key}' from the global DSO configurations...")
            parent = get_dict_item(self.global_config_, key.split('.')[:-1])
            if parent and key.split('.')[-1] in parent:
                parent.pop(key.split('.')[-1])
                self.save_global_config()
                self.load_global_config()
            else:
                raise DSOException(f"'{key}' not found in the global DSO configuratoins.")

        else:
            Logger.info(f"Deleting '{key}' from the local DSO configurations...")
            parent = get_dict_item(self.local_config_raw, key.split('.')[:-1])
            if parent and key.split('.')[-1] in parent:
                parent.pop(key.split('.')[-1])
                self.save_local_config()
                self.load_local_config()
            else:
                raise DSOException(f"'{key}' not found in the local DSO configuratoins.")

    def init(self, working_dir, init_config, config_overrides, use_local):
        Logger.info("Initializing DSO configurations...")
        self.working_dir = working_dir

        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)

        if os.path.exists(self.local_config_file_path):
            Logger.warn("The working directory has already been initialized.")

        config = {}
        
        ### use init_config instead of local/inherited config
        if init_config:
            config = merge_dicts(init_config, config)
            ### merge with existing local configuration?
            if use_local:
                Logger.debug("Merging local configuration...")
                self.load_local_config(True)
                config = merge_dicts(self.local_config, config)
        else:
            ### do not show warning if directory is not initialized yet
            self.load_local_config(True)
            config = merge_dicts(self.local_config, config)
            ### override locally inherited configuration?
            if use_local: 
                Logger.debug("Merging global configuration...")
                self.load_global_config()
                config = merge_dicts(self.global_config, config)

                Logger.debug("Merging inherited configuration...")
                self.load_inherited_config()
                config = merge_dicts(self.inherited_config, config)

        ### if config overrides, merge them to local
        if config_overrides:
            Logger.debug("Merging configuration overrides...")
            self.apply_config_overrides(config_overrides)
            config = merge_dicts(self.overriden_config, config)

        self.local_config = config
        self.update_merged_config()

        self.save_local_config()


    def parse_path(self, path):
        """
            path is in the form of [/]namespace/project/application/stage/env_no/[key]
        """
        parts = path.split('/')
        if not parts[0]: parts.pop(0)
        namespace = parts[0]
        project = parts[1]
        application = parts[2]
        stage = f"{parts[3]}/{parts[4]}"
        key = '/'.join(parts[5:]) if len(parts) > 5 else None
        return namespace, project, application, stage, key


    def get_hierachy_paths(self, context, key=None, path_prefix='', ignore_stage=False, uninherited=False, reverse=False):
        
        result = []
        if uninherited:
            result.append(path_prefix + context.get_path(key=key))
        else:
            ### Add the global context: /default/default/default/0
            result.append(path_prefix + Context(context.namespace, 'default', 'default', 'default/0').get_path(key))
            if not ignore_stage and not Stages.is_default(context.stage):
                ### Add global stage context
                result.append(path_prefix + Context(context.namespace, 'default', 'default', Stages.get_default_env(context.stage)).get_path(key))
                ### Add global numbered stage context
                if not Stages.is_default_env(context.stage):
                    result.append(path_prefix + Context(context.namespace, 'default', 'default', stage).get_path(key))

            if not context.project == 'default':
                ### Add the project context: /project/default/default/0
                result.append(path_prefix + Context(context.namespace, context.project, 'default', 'default/0').get_path(key))
                if not ignore_stage and not Stages.is_default(context.stage):
                    ### Add the project stage context: /project/default/stage/0
                    result.append(path_prefix + Context(context.namespace, context.project, 'default', Stages.get_default_env(context.stage)).get_path(key))
                    ### Add the project numbered stage context: /project/default/stage/env
                    if not Stages.is_default_env(context.stage):
                        result.append(path_prefix + Context(context.namespace, context.project, 'default', stage).get_path(key))
                
                if not context.application == 'default':
                    ### Add the application context: /project/application/default/0
                    result.append(path_prefix + Context(context.namespace, context.project, context.application, 'default/0').get_path(key))
                    if not ignore_stage and not Stages.is_default(context.stage):
                        ### Add the project stage context: /project/application/stage/0
                        result.append(path_prefix + Context(context.namespace, context.project, context.application, Stages.get_default_env(context.stage)).get_path(key))
                        ### Add the application numbered stage context: /dso/project/application/stage/env
                        if not Stages.is_default_env(context.stage):
                            result.append(path_prefix + Context(context.namespace, context.project, context.application, context.stage).get_path(key))

        return list(reversed(result)) if reverse else result



Contexts = ContextService()